"""
class that specifies where we read the cloudmesh.yaml file from
"""
import os

from cloudmesh.common.Shell import Shell
from cloudmesh.common.util import path_expand

__config_dir_prefix__ = os.path.join("~", ".cloudmesh")

__config_dir__ = path_expand(__config_dir_prefix__)


def config_file(filename):
    """
    The location of the config file: ~/.cloudmesh/filename. ~ will be expanded
    :param filename: the filename
    """
    return os.path.join(__config_dir__, filename)


def config_file_raw(filename):
    """
    The location of the config file: ~/.cloudmesh/filename. ~ will NOT be expanded
    :param filename: the filename
    """
    return os.path.join(__config_dir_prefix__, filename)


def config_file_prefix():
    """
    The prefix of the configuration file location
    """
    return __config_dir_prefix__


def config_dir_setup(filename):
    """
    sets the config file and makes sure the directory exists if it has not yet been created.
    :param filename: 
    :return: 
    """
    path = os.path.dirname(filename)
    if not os.path.isdir(path):
        Shell.mkdir(path)




"""
Useful utility functions
"""
from __future__ import print_function

import collections
import glob
import inspect
import os
import random
import re
import shutil
import tempfile
# import pip
import time
from contextlib import contextmanager
from string import Template

from six.moves import input


@contextmanager
def tempdir(*args, **kwargs):
    """A contextmanager to work in an auto-removed temporary directory

    Arguments are passed through to tempfile.mkdtemp

    example:

    >>> with tempdir() as path:
    ...    pass
    """

    d = tempfile.mkdtemp(*args, **kwargs)
    try:
        yield d
    finally:
        shutil.rmtree(d)


def exponential_backoff(fn, sleeptime_s_max=30 * 60):
    """Calls `fn` until it returns True, with an exponentially increasing wait time between calls"""

    sleeptime_ms = 500
    while True:
        if fn():
            return True
        else:
            print('Sleeping {} ms'.format(sleeptime_ms))
            time.sleep(sleeptime_ms / 1000.0)
            sleeptime_ms *= 2

        if sleeptime_ms / 1000.0 > sleeptime_s_max:
            return False


def search(lines, pattern):
    """
    return all lines that match the pattern
    #TODO: we need an example
    
    :param lines: 
    :param pattern: 
    :return: 
    """
    p = pattern.replace("*", ".*")
    test = re.compile(p)
    result = []
    for l in lines:
        if test.search(l):
            result.append(l)
    return result


def grep(pattern, filename):
    """Very simple grep that returns the first matching line in a file.
    String matching only, does not do REs as currently implemented.
    """
    try:
        # for line in file
        # if line matches pattern:
        #    return line
        return next((L for L in open(filename) if L.find(pattern) >= 0))
    except StopIteration:
        return ''


def path_expand(text):
    """ returns a string with expanded variable.

    :param text: the path to be expanded, which can include ~ and $ variables
    :param text: string

    """
    result = os.path.expanduser(text)

    # template = Template(text)
    # result = template.substitute(os.environ)

    if result.startswith("."):
        result = result.replace(".", os.getcwd(), 1)
    return result


def convert_from_unicode(data):
    """
    converts unicode data to a string
    :param data: the data to convert
    :return: 
    """
    # if isinstance(data, basestring):

    if isinstance(data, str):
        return str(data)
    elif isinstance(data, collections.Mapping):
        return dict(map(convert_from_unicode, data.items()))
    elif isinstance(data, collections.Iterable):
        return type(data)(map(convert_from_unicode, data))
    else:
        return data


def yn_choice(message, default='y', tries=None):
    """asks for a yes/no question.
    :param message: the message containing the question
    :param default: the default answer
    """
    # http://stackoverflow.com/questions/3041986/python-command-line-yes-no-input"""
    choices = 'Y/n' if default.lower() in ('y', 'yes') else 'y/N'
    if tries is None:
        choice = input("%s (%s) " % (message, choices))
        values = ('y', 'yes', '') if default == 'y' else ('y', 'yes')
        return True if choice.strip().lower() in values else False
    else:
        while tries > 0:
            choice = input("%s (%s) (%s)" % (message, choices, "'q' to discard"))
            choice = choice.strip().lower()
            if choice in ['y', 'yes']:
                return True
            elif choice in ['n', 'no', 'q']:
                return False
            else:
                print("Invalid input...")
                tries -= 1


def banner(txt=None, c="#", debug=True):
    """prints a banner of the form with a frame of # around the txt::

      ############################
      # txt
      ############################

    .

    :param txt: a text message to be printed
    :type txt: string
    :param c: the character used instead of c
    :type c: character
    """
    if debug:
        print()
        print("#", 70 * c)
        if txt is not None:
            print("#", txt)
            print("#", 70 * c)


def str_banner(txt=None, c="#", debug=True):
    """prints a banner of the form with a frame of # around the txt::

      ############################
      # txt
      ############################

    .

    :param txt: a text message to be printed
    :type txt: string
    :param c: the character used instead of c
    :type c: character
    """
    line = ""
    if debug:
        line += "\n"
        line += "# " + str(70 * c)
        if txt is not None:
            line += "# " + txt
            line += "# " + str(70 * c)
    return line


# noinspection PyPep8Naming
def HEADING(txt=None):
    """
    Prints a message to stdout with #### surrounding it. This is useful for
    nosetests to better distinguish them.

    :param txt: a text message to be printed
    :type txt: string
    """
    frame = inspect.getouterframes(inspect.currentframe())

    filename = frame[1][1].replace(os.getcwd(), "")
    line = frame[1][2] - 1
    method = frame[1][3]
    msg = "{}\n# {} {} {}".format(txt, method, filename, line)

    print()
    banner(msg)


def backup_name(filename):
    """
    :param filename: given a filename creates a backup name of the form
                     filename.bak.1. If the filename already exists
                     the number will be increased as  much as needed so
                     the file does not exist in the given location.
                     The filename can consists a path and is expanded
                     with ~ and environment variables.
    :type filename: string
    :rtype: string
    """
    location = path_expand(filename)
    n = 0
    found = True
    backup = None
    while found:
        n += 1
        backup = "{0}.bak.{1}".format(location, n)
        found = os.path.isfile(backup)
    return backup


def auto_create_version(class_name, version, filename="__init__.py"):
    """
    creates a version number in the __init__.py file.
    it can be accessed with __version__
    :param class_name: 
    :param version: 
    :param filename: 
    :return: 
    """
    version_filename = "{0}/{1}".format(class_name, filename)
    with open(version_filename, "r") as f:
        content = f.read()

    if content != '__version__ = "{0}"'.format(version):
        banner("Updating version to {0}".format(version))
        with open(version_filename, "w") as text_file:
            text_file.write('__version__ = "{0:s}"'.format(version))


def auto_create_requirements(requirements):
    """
    creates a requirement.txt file form the requirements in the list. If the file exists, it get changed only if the
    requirements in the list are different from the existing file

    :param requirements: the requirements in a list
    """
    banner("Creating requirements.txt file")
    try:
        with open("requirements.txt", "r") as f:
            file_content = f.read()
    except:
        file_content = ""

    setup_requirements = '\n'.join(requirements)

    if setup_requirements != file_content:
        with open("requirements.txt", "w") as text_file:
            text_file.write(setup_requirements)


def copy_files(files_glob, source_dir, dest_dir):
    """

    :param files_glob: *.yaml
    :param source_dir: source directiry
    :param dest_dir: destination directory
    :return:
    """
    files = glob.iglob(os.path.join(source_dir, files_glob))
    for filename in files:
        if os.path.isfile(filename):
            shutil.copy2(filename, dest_dir)


def dict_replace(content, replacements=None):
    # TODO: it is unclear why we have this. Describe
    if replacements is None:
        replacements = {}
    for key in replacements:
        content = content.replace("\{key\}".format(replacements[key]))

    return content


def readfile(filename):
    """
    returns the content of a file
    :param filename: the filename
    :return: 
    """
    with open(path_expand(filename), 'r') as f:
        content = f.read()
    return content


def writefile(filename, content):
    """
    writes the content into the file
    :param filename: the filename
    :param content: teh content
    :return: 
    """
    outfile = open(path_expand(filename), 'w')
    outfile.write(content)
    outfile.close()


# Reference: http://interactivepython.org/runestone/static/everyday/2013/01/3_password.html
def generate_password(length=8, lower=True, upper=True, number=True):
    """
    generates a simple password. We should not realy use this in production.
    :param length: the length of the password
    :param lower: True of lower case characters are allowed
    :param upper: True if upper case characters are allowed
    :param number: True if numbers are allowed
    :return: 
    """
    lletters = "abcdefghijklmnopqrstuvwxyz"
    uletters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    # This doesn't guarantee both lower and upper cases will show up
    alphabet = lletters + uletters
    digit = "0123456789"
    mypw = ""

    def _random_character(texts):
        return texts[random.randrange(len(texts))]

    if not lower:
        alphabet = uletters
    elif not upper:
        alphabet = lletters

    for i in range(length):
        # last half length will be filled with numbers
        if number and i >= int(length / 2):
            mypw = mypw + _random_character(digit)
        else:
            mypw = mypw + _random_character(alphabet)
    return mypw



"""
A simple framework to handle error messages 
"""
import sys
import traceback


#
# TODO: this class seems to replicate some portions of what Console does
#
class Error(object):
    """
    A class to print error messages
    """

    #
    # TODO: this should probably use Console so we can print in color
    #
    @classmethod
    def msg(cls, error=None, debug=True, trace=True):
        """
        prints the error message
        :param error: the error message
        :param debug: only prints it if debug is set to true
        :param trace: if true prints the trace
        :return: 
        """
        if debug and error is not None:
            print(error)
        # TODO: BUG: trace should only be printed if debug is true
        if trace:
            print(traceback.format_exc())

    @classmethod
    def traceback(cls, error=None, debug=True, trace=True):
        """
        prints the trace
        :param error: a message preceding the trace
        :param debug: prints it if debug is set to true
        :param trace: 
        :return: 
        """
        # TODO: if debug:
        Error.msg(error=error, debug=debug, trace=trace)

    @classmethod
    def info(cls, msg):
        """
        prints an info msg. 
        :param msg: the message
        :return: 
        """
        # TODO: if info:
        print(msg)

    @classmethod
    def warning(cls, msg):
        """
        prints a warning message. 
        :param msg: 
        :return: 
        """
        # TODO: if warning:
        print(msg)

    @classmethod
    def debug(cls, msg):
        """
        prints a debug message.
        :param msg: the message
        :return: 
        """
        # TODO: if debug:
        print(msg)

    @classmethod
    def exit(cls, msg):
        """
        call a system exit
        :param msg: 
        :return: 
        """
        # TODO: if debug:
        print(msg)
        sys.exit()

        # TODO: adding methods for setting and getting info and debug are missing



from cloudmesh.common.dotdict import dotdict
from cloudmesh.common.util import path_expand
from cloudmesh.common.util import readfile
from ruamel import yaml


class Config(object):
    def __init__(self, filename=None):
        self.filename = filename
        if self.filename is None:
            self.filename = path_expand("~/.cloudmesh/cloudmesh2.yaml")
        content = readfile(self.filename)
        self.data = yaml.load(content, Loader=yaml.RoundTripLoader)

    def cloud(self, name):
        return self.data["cloudmesh"]["clouds"][name]

    def credentials(self, name):
        return dotdict(self.cloud(name)["credentials"])


'''
if __name__ == "__main__":
    from pprint import pprint
    c = Config()
    pprint(c.credentials("jet"))
'''



"""
A convenient method to execute shell commands and return their output. Note: that this method requires that the
command be completely execute before the output is returned. FOr many activities in cloudmesh this is sufficient.
"""

from __future__ import print_function

import errno
import glob
import os
import subprocess
import sys
import zipfile
from pipes import quote
from sys import platform

from cloudmesh.common.console import Console
from cloudmesh.common.util import path_expand
from distutils.spawn import find_executable


class Brew(object):
    @classmethod
    def install(cls, name):

        r = Shell.brew("install", name)
        print(r)

        if "already satisfied: " + name in r:
            print(name, "... already installed")
            Console.ok(r)
        elif "Successfully installed esptool":
            print(name, "... installed")
            Console.ok(r)
        else:
            print(name, "... error")
            Console.error(r)

    @classmethod
    def version(cls, name):
        r = Shell.brew("ls", "--version", "name")
        print(r)


class Pip(object):
    @classmethod
    def install(cls, name):
        r = Shell.pip("install", "esptool")
        if "already satisfied: esptool" in r:
            print(name, "... already installed")
            Console.ok(r)
        elif "Successfully installed esptool":
            print(name, "... installed")
            Console.ok(r)
        else:
            print(name, "... error")
            Console.error(r)


class SubprocessError(Exception):
    """
    Manages the formatting of the error and stdout. 
    THis command should not be directly called. Instead use SHell
    """

    def __init__(self, cmd, returncode, stderr, stdout):
        """
        sets the error
        :param cmd: the command executed
        :param returncode: the return code
        :param stderr: the stderr
        :param stdout: the stdout
        """
        self.cmd = cmd
        self.returncode = returncode
        self.stderr = stderr
        self.stdout = stdout

    def __str__(self):

        def indent(lines, amount, ch=' '):
            """
            indent the lines by multiples of ch
            :param lines: 
            :param amount: 
            :param ch: 
            :return: 
            """
            padding = amount * ch
            return padding + ('\n' + padding).join(lines.split('\n'))

        cmd = ' '.join(map(quote, self.cmd))
        s = ''
        s += 'Command: %s\n' % cmd
        s += 'Exit code: %s\n' % self.returncode

        if self.stderr:
            s += 'Stderr:\n' + indent(self.stderr, 4)
        if self.stdout:
            s += 'Stdout:\n' + indent(self.stdout, 4)

        return s


class Subprocess(object):
    """
    Executes a command. This class should not be directly used, but instead you should use Shell.
    """

    def __init__(self, cmd, cwd=None, stderr=subprocess.PIPE, stdout=subprocess.PIPE, env=None):
        """
        execute the given command
        :param cmd: the command
        :param cwd: the directory in which to execute the command
        :param stderr: the pipe for stderror
        :param stdout: the pipe for the stdoutput
        :param env: 
        """
        Console.debug_msg('Running cmd: {}'.format(' '.join(map(quote, cmd))))

        proc = subprocess.Popen(cmd, stderr=stderr, stdout=stdout, cwd=cwd, env=env)
        stdout, stderr = proc.communicate()

        self.returncode = proc.returncode
        self.stderr = stderr
        self.stdout = stdout

        if self.returncode != 0:
            raise SubprocessError(cmd, self.returncode, self.stderr, self.stdout)


class Shell(object):
    """
    The shell class allowing us to conveniently access many operating system commands. 
    TODO: This works well on Linux and OSX, but has not been tested much on Windows
    """

    # TODO: we have not supported cygwin for a while
    cygwin_path = 'bin'  # i copied fom C:\cygwin\bin

    command = {
        'windows': {},
        'linux': {},
        'darwin': {}
    }

    '''
    TODO
    
    how do we now define dynamically functions based on a list that we want to support

    what we want is where args are multiple unlimited parameters to the function

    def f(args...):
        name = get the name from f
        a = list of args...

        cls.execute(cmd, arguments=a, capture=True, verbose=False)

    commands = ['ps', 'ls', ..... ]
    for c in commands:
        generate this command and add to this class dynamically

    or do something more simple

    ls = cls.execute('cmd', args...)

    '''

    @classmethod
    def get_python(cls):
        """
        returns the python and pip version
        :return: python version, pip version
        """
        python_version = sys.version_info[:3]
        v_string = [str(i) for i in python_version]
        python_version_s = '.'.join(v_string)

        # pip_version = pip.__version__
        pip_version = Shell.pip("--version").split()[1]
        return python_version_s, pip_version

    @classmethod
    def check_python(cls):
        """
        checks if the python version is supported
        :return: True if it is supported
        """
        python_version = sys.version_info[:3]

        v_string = [str(i) for i in python_version]

        python_version_s = '.'.join(v_string)
        if (python_version[0] == 2) and (python_version[1] >= 7) and (python_version[2] >= 9):

            print("You are running a supported version of python: {:}".format(python_version_s))
        else:
            print("WARNING: You are running an unsupported version of python: {:}".format(python_version_s))
            print("         We recommend you update your python")

        # pip_version = pip.__version__
        python_version, pip_version = cls.get_python()

        if int(pip_version.split(".")[0]) >= 9:
            print("You are running a supported version of pip: " + str(pip_version))
        else:
            print("WARNING: You are running an old version of pip: " + str(pip_version))
            print("         We recommend you update your pip  with \n")
            print("             pip install -U pip\n")

    @classmethod
    def check_output(cls, *args, **kwargs):
        """Thin wrapper around :func:`subprocess.check_output`
        """
        return subprocess.check_output(*args, **kwargs)

    @classmethod
    def ls(cls, *args):
        """
        executes ls with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('ls', args)

    @classmethod
    def ps(cls, *args):
        """
        executes ps with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('ps', args)

    @classmethod
    def bash(cls, *args):
        """
        executes bash with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('bash', args)

    @classmethod
    def brew(cls, *args):
        """
        executes bash with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('brew', args)

    @classmethod
    def cat(cls, *args):
        """
        executes cat with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('cat', args)

    @classmethod
    def git(cls, *args):
        """
        executes git with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('git', args)

    # noinspection PyPep8Naming
    @classmethod
    def VBoxManage(cls, *args):
        """
        executes VboxManage with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('VBoxManage', args)

    @classmethod
    def blockdiag(cls, *args):
        """
        executes blockdiag with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('blockdiag', args)

    @classmethod
    def cm(cls, *args):
        """
        executes cm with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('cm', args)

    @classmethod
    def fgrep(cls, *args):
        """
        executes fgrep with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('fgrep', args)

    @classmethod
    def grep(cls, *args):
        """
        executes grep with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('grep', args)

    @classmethod
    def head(cls, *args):
        """
        executes head with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('head', args)

    @classmethod
    def keystone(cls, *args):
        """
        executes keystone with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('keystone', args)

    @classmethod
    def kill(cls, *args):
        """
        executes kill with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('kill', args)

    @classmethod
    def nosetests(cls, *args):
        """
        executes nosetests with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('nosetests', args)

    @classmethod
    def nova(cls, *args):
        """
        executes nova with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('nova', args)

    @classmethod
    def ping(cls, host=None, count=1):
        """
        execute ping
        :param host: the host to ping 
        :param count: the number of pings
        :return: 
        """
        return cls.execute('ping', "-c {} {}".format(count, host))

    @classmethod
    def pwd(cls, *args):
        """
        executes pwd with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('pwd', args)

    @classmethod
    def rackdiag(cls, *args):
        """
        executes rackdiag with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('rackdiag', args)

    @classmethod
    def rm(cls, *args):
        """
        executes rm with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('rm', args)

    @classmethod
    def rsync(cls, *args):
        """
        executes rsync with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('rsync', args)

    @classmethod
    def scp(cls, *args):
        """
        executes scp with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('scp', args)

    @classmethod
    def sort(cls, *args):
        """
        executes sort with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('sort', args)

    @classmethod
    def sh(cls, *args):
        """
        executes sh with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('sh', args)

    @classmethod
    def ssh(cls, *args):
        """
        executes ssh with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('ssh', args)

    @classmethod
    def sudo(cls, *args):
        """
        executes sudo with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('sudo', args)

    @classmethod
    def tail(cls, *args):
        """
        executes tail with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('tail', args)

    @classmethod
    def vagrant(cls, *args):
        """
        executes vagrant with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('vagrant', args)

    @classmethod
    def pandoc(cls, *args):
        """
        executes vagrant with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('pandoc', args)

    @classmethod
    def mongod(cls, *args):
        """
        executes mongod with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('mongod', args)

    @classmethod
    def dialog(cls, *args):
        """
        executes dialof with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('dialog', args)

    @classmethod
    def pip(cls, *args):
        """
        executes pip with the given arguments
        :param args: 
        :return: 
        """
        return cls.execute('pip', args)

    @classmethod
    def remove_line_with(cls, lines, what):
        """
        returns all lines that do not contain what
        :param lines: 
        :param what: 
        :return: 
        """
        result = []
        for line in lines:
            if what not in line:
                result = result + [line]
        return result

    @classmethod
    def find_lines_with(cls, lines, what):
        """
        returns all lines that contain what
        :param lines: 
        :param what: 
        :return: 
        """
        result = []
        for line in lines:
            if what in line:
                result = result + [line]
        return result

    def __init__(cls):
        """
        identifies parameters for the os
        """
        if cls.operating_system() == "windows":
            cls.find_cygwin_executables()
        else:
            pass
            # implement for cmd, for linux we can just pass as it includes everything

    @classmethod
    def find_cygwin_executables(cls):
        """
        find the executables in cygwin
        """
        exe_paths = glob.glob(cls.cygwin_path + r'\*.exe')
        # print cls.cygwin_path
        # list all *.exe in  cygwin path, use glob
        for c in exe_paths:
            exe = c.split('\\')
            name = exe[1].split('.')[0]
            # command['windows'][name] = "{:}\{:}.exe".format(cygwin_path, c)
            cls.command['windows'][name] = c

    @classmethod
    def terminal_type(cls):
        """
        returns  darwin, cygwin, cmd, or linux
        """
        what = sys.platform

        kind = 'UNDEFINED_TERMINAL_TYPE'
        if 'linux' in what:
            kind = 'linux'
        elif 'darwin' in what:
            kind = 'darwin'
        elif 'cygwin' in what:
            kind = 'cygwin'
        elif 'windows' in what:
            kind = 'windows'

        return kind

    @classmethod
    def which(cls, command):
        """
        returns the path of the command with which
        :param command: teh command
        :return: the path
        """
        return find_executable(command)

    @classmethod
    def command_exists(cls, name):
        """
        returns True if the command exists
        :param name: 
        :return: 
        """
        return cls.which(name) is not None

    @classmethod
    def operating_system(cls):
        """
        the name of the os
        :return: the name of the os
        """
        return platform.system().lower()

    @classmethod
    def execute(cls, cmd, arguments="", shell=False, cwd=None, traceflag=True, witherror=True):
        """Run Shell command

        :param witherror: if set to fasle the error will not be printed
        :param traceflag: if set to true the trace is printed in case of an error
        :param cwd: the current working directory in whcih the command is supposed to be executed.
        :param shell: if set to true the subprocess is called as part of a shell
        :param cmd: command to run
        :param arguments: we do not know yet
        :return:
        """
        # print "--------------"
        result = None
        terminal = cls.terminal_type()
        # print cls.command
        os_command = [cmd]
        if terminal in ['linux', 'windows']:
            os_command = [cmd]
        elif 'cygwin' in terminal:
            if not cls.command_exists(cmd):
                print("ERROR: the command could not be found", cmd)
                return
            else:
                os_command = [cls.command[cls.operating_system()][cmd]]

        if isinstance(arguments, list):
            os_command = os_command + arguments
        elif isinstance(arguments, tuple):
            os_command = os_command + list(arguments)
        elif isinstance(arguments, str):
            os_command = os_command + arguments.split()
        else:
            print("ERROR: Wrong parameter type", type(arguments))

        if cwd is None:
            cwd = os.getcwd()
        try:
            if shell:
                result = subprocess.check_output(
                    os_command,
                    stderr=subprocess.STDOUT,
                    shell=True,
                    cwd=cwd)
            else:
                result = subprocess.check_output(
                    os_command,
                    # shell=True,
                    stderr=subprocess.STDOUT,
                    cwd=cwd)
        except:
            if witherror:
                Console.error("problem executing subprocess", traceflag=traceflag)
        if result is not None:
            result = result.strip().decode()
        return result

    @classmethod
    def mkdir(cls, directory):
        """
        creates a directory with all its parents in ots name
        :param directory: the path of the directory
        :return: 
        """
        directory = path_expand(directory)
        try:
            os.makedirs(directory)
        except OSError as e:

            # EEXIST (errno 17) occurs under two conditions when the path exists:
            # - it is a file
            # - it is a directory
            #
            # if it is a file, this is a valid error, otherwise, all
            # is fine.
            if e.errno == errno.EEXIST and os.path.isdir(directory):
                pass
            else:
                raise

    def unzip(cls, source_filename, dest_dir):
        """
        unzips a file into the destination directory
        :param dest_dir: the destination directory
        :return:
        """

        with zipfile.ZipFile(source_filename) as zf:
            for member in zf.infolist():
                # Path traversal defense copied from
                # http://hg.python.org/cpython/file/tip/Lib/http/server.py#l789
                words = member.filename.split('/')
                path = path_expand(dest_dir)
                for word in words[:-1]:
                    drive, word = os.path.splitdrive(word)
                    head, word = os.path.split(word)
                    if word in (os.curdir, os.pardir, ''):
                        continue
                    path = os.path.join(path, word)
                zf.extract(member, path)


def main():
    """
    a test that should actually be added into a nosetest
    :return: 
    """
    shell = Shell()

    print(shell.terminal_type())

    r = shell.execute('pwd')  # copy line replace
    print(r)

    # shell.list()

    # print json.dumps(shell.command, indent=4)

    # test some commands without args
    """
    for cmd in ['whoami', 'pwd']:
        r = shell._execute(cmd)
        print ("---------------------")
        print ("Command: {:}".format(cmd))
        print ("{:}".format(r))
        print ("---------------------")
    """
    r = shell.execute('ls', ["-l", "-a"])
    print(r)

    r = shell.execute('ls', "-l -a")
    print(r)

    r = shell.ls("-aux")
    print(r)

    r = shell.ls("-a", "-u", "-x")
    print(r)

    r = shell.pwd()
    print(r)


if __name__ == "__main__":
    main()



"""
A convenient dot dict class

a = dotdict({"argument": "value"})

print (a.argument)

"""


# noinspection PyPep8Naming
class dotdict(dict):
    """dot.notation access to dictionary attributes"""

    def __getattr__(self, attr):
        return self.get(attr)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__



###############################################################
#  tests

#
# TODO: theres is absolutely no need to use hypothesis. THis code does not
#       follow our convention to develop nose tests. Also usining nosetests allows us
#       to get rid of hypothesis
#
import unittest
from hypothesis import given
from hypothesis import strategies as st
from tempfile import mkstemp

# TODO strdb import missing

class TestYamlDB(unittest.TestCase):
    def setUp(self):
        self.dbfd, self.dbpath = mkstemp(
            prefix='cloudmesh.db.strdb.TestYamlDB.'
        )
        self.db = YamlDB(path=self.dbpath)

    def tearDown(self):
        os.close(self.dbfd)
        os.unlink(self.dbpath)

    @given(st.text(), st.text())
    def test_insert(self, key, value):
        self.db[key] = value

    @given(st.text(), st.text())
    def test_in(self, key, value):
        self.db[key] = value
        self.assertIn(key, self.db)

    @given(st.text(), st.text())
    def test_get(self, key, value):
        self.db[key] = value
        dbvalue = self.db[key]
        self.assertEqual(value, dbvalue)

    @given(st.text(), st.text())
    def test_del(self, key, value):
        self.db[key] = value
        del self.db[key]
        self.assertNotIn(key, self.db)

    @given(st.text(), st.text())
    def test_persistence(self, key, value):
        self.db[key] = value
        with open(self.dbpath, 'rb') as dbfile:
            db = yaml.safe_load(dbfile)
        self.assertIn(key, db)
        self.assertEqual(value, db[key])

    @given(st.text(), st.text())
    def test_iter(self, key, value):
        self.db[key] = value
        for k in self.db:
            pass

    @given(st.text(), st.text())
    def test_clear(self, key, value):
        self.db[key] = value
        self.db.clear()
        self.assertNotIn(key, self.db)

    @given(st.lists(st.tuples(st.text(), st.text())))
    def test_len(self, entries):
        self.db.clear()

        duplicates = 0
        for k, v in entries:
            duplicates += 1 if k in self.db else 0
            self.db[k] = v
        self.assertEqual(len(self.db), len(entries) - duplicates)


if __name__ == '__main__':
    unittest.main()


"""
simple logging convenience framework
"""
import logging
import os

from cloudmesh.common.locations import config_file
from cloudmesh.common.util import grep


# noinspection PyUnusedLocal,PyPep8Naming
def LOGGER(filename):
    """creates a logger with the given name.

    You can use it as follows::

       log = cloudmesh.util.LOGGER(__file__)
       log.error("this is an error")
       log.info("this is an info")
       log.warning("this is a warning")

    """
    pwd = os.getcwd()
    name = filename.replace(pwd, "$PWD")
    try:
        (first, name) = name.split("site-packages")
        name += "... site"
    except:
        pass

    loglevel = logging.CRITICAL
    try:
        level = grep("loglevel:", config_file(
            "/cloudmesh_debug.yaml")).strip().split(":")[1].strip().lower()

        if level.upper() == "DEBUG":
            loglevel = logging.DEBUG
        elif level.upper() == "INFO":
            loglevel = logging.INFO
        elif level.upper() == "WARNING":
            loglevel = logging.WARNING
        elif level.upper() == "ERROR":
            loglevel = logging.ERROR
        else:
            level = logging.CRITICAL
    except:
        # print "LOGLEVEL NOT FOUND"
        loglevel = logging.DEBUG

    log = logging.getLogger(name)
    log.setLevel(loglevel)

    formatter = logging.Formatter(
        'CM {0:>50}:%(lineno)s: %(levelname)6s - %(message)s'.format(name))

    # formatter = logging.Formatter(
    #    'CM {0:>50}: %(levelname)6s - %(module)s:%(lineno)s %funcName)s: %(message)s'.format(name))
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    log.addHandler(handler)
    return log


# noinspection PyPep8Naming
def LOGGING_ON(log):
    """
    Switches logging on
    :param log: the logger for which we switch logging on 
    """
    try:
        log.setLevel(logging.DEBUG)
        return True
    except:
        return False


# noinspection PyPep8Naming
def LOGGING_OFF(log):
    """
    Switches logging off
    :param log: the logger for which we switch logging off
    """
    try:
        log.setLevel(logging.CRITICAL)
        return True
    except:
        return False



"""
Managing ~/.ssh/config
"""
from __future__ import print_function

import json
import os
from textwrap import dedent

from cloudmesh.shell.console import Console

from cloudmesh.common.Shell import Shell


# noinspection PyPep8Naming
class ssh_config(object):
    """
    Managing the config in .ssh
    """
    def __init__(self, filename=None):
        if filename is not None:
            # load
            pass
        else:
            filename = "~/.ssh/config"

        self.filename = os.path.expanduser(filename)
        self.load()

    def names(self):
        """
        The names defined in ~/.ssh/config
        :return: the names
        """
        found_names = []
        with open(self.filename) as f:
            content = f.readlines()
        for line in content:
            line = line.strip()
            if " " in line:
                attribute, value = line.split(" ", 1)
                attribute = attribute.strip()
                value = value.strip()
                if attribute.lower() in ['host']:
                    found_names.append(value)
        return found_names

    # noinspection PyAttributeOutsideInit
    def load(self):
        """list the hosts defined in the ssh config file"""
        with open(self.filename) as f:
            content = f.readlines()
        content = [" ".join(x.split()).strip('\n').lstrip().split(' ', 1) for x in content]

        # removes duplicated spaces, and splits in two fields, removes leading spaces
        hosts = {}
        host = "NA"
        for line in content:
            if line[0].startswith('#') or line[0] is '':
                pass  # ignore line
            else:
                attribute = line[0]
                value = line[1]
                if attribute.lower().strip() in ['Host', 'host']:
                    host = value
                    hosts[host] = {'host': host}
                else:
                    # In case of special configuration lines, such as port
                    # forwarding,
                    # there would be no 'Host india' line.
                    if host in hosts:
                        hosts[host][attribute] = value
                        # pass
        self.hosts = hosts

    def list(self):
        """
        list the hosts in the config file
        :return: 
        """
        return list(self.hosts.keys())

    def __str__(self):
        """
        The string representing the config file
        :return: the string representation
        """
        return json.dumps(self.hosts, indent=4)

    def status(self):
        """
        executes a test with the given ssh config if a login is possible. 
           TODO: not yet implemented
        """
        pass

    def login(self, name):
        """
        login to the host defines in .ssh/config by name
        :param name: the name of the host as defined in the config file
        :return: 
        """
        os.system("ssh {0}".format(name))

    def execute(self, name, command):
        """
        execute the command on the named host
        :param name: the name of the host in config
        :param command: the command to be executed
        :return: 
        """
        if name in ["localhost"]:
            r = '\n'.join(Shell.sh("-c", command).split()[-1:])
        else:
            r = '\n'.join(Shell.ssh(name, command).split()[-1:])
        return r

    def local(self, command):
        """
        execute the command on the localhost
        :param command: the command to execute
        :return: 
        """
        return self.execute("localhost", command)

    def username(self, host):
        """
        returns the username for a given host in the config file
        :param host: the hostname
        :return: the username
        """
        if host in self.hosts:
            return self.hosts[host]["User"]
        else:
            return None

    def generate(self,
                 key="india",
                 host="india.futuresystems.org",
                 username=None,
                 force=False,
                 verbose=False):
        """
        adds a host to the config file with given parameters.  #TODO: make sure this is better documented
        :param key: the key
        :param host: the host
        :param username: the username
        :param force: not used
        :param verbose: prints debug messages
        :return: 
        """
        data = {
            "host": host,
            "key": key,
            "username": username
        }
        if verbose and key in self.names():
            Console.error("{key} already in ~/.ssh/config".format(**data), traceflag=False)
            return ""
        else:
            entry = dedent("""
            Host {key}
                Hostname {host}
                User {username}
            """.format(**data))
        try:
            with open(self.filename, "a") as config_ssh:
                config_ssh.write(entry)
            config_ssh.close()
            self.load()
            if verbose:
                Console.ok("Added india to ~/.ssh/config")
        except Exception as e:
            if verbose:
                Console.error(e.message)


if __name__ == "__main__":
    from cloudmesh.common.ConfigDict import ConfigDict

    hosts = ssh_config()

    user = ConfigDict("cloudmesh.yaml")["cloudmesh.profile.user"]
    print("User:", user)

    hosts.generate(key="india", username=user)
    print(hosts.filename)

    print(hosts.list())
    print(hosts)

    import sys

    sys.exit()

    r = hosts.execute("india", "hostname")
    print(r)

    r = hosts.execute("localhost", "hostname")
    print(r)

    r = hosts.local("hostname")
    print(r)

    # hosts.login("india")



"""
Manage the cloudmesh.yaml file
"""
from __future__ import print_function

import json
import os.path
import sys
from collections import OrderedDict

from ruamel import yaml

from cloudmesh.common.BaseConfigDict import BaseConfigDict
from cloudmesh.common.console import Console
from cloudmesh.common.todo import TODO
from cloudmesh.common.util import backup_name
from cloudmesh.common.util import path_expand


def custom_print(data_structure, indent, attribute_indent=4):
    """
    prints the data structure at a given level. This includes dicts and ordered dicts
    :param data_structure: 
    :param indent: 
    :param attribute_indent: 
    :return: 
    """
    for key, value in data_structure.items():
        print("\n%s%s:" % (' ' * attribute_indent * indent, str(key)), end=' ')
        if isinstance(value, OrderedDict):
            custom_print(value, indent + 1)
        elif isinstance(value, dict):
            custom_print(value, indent + 1)
        else:
            print("%s" % (str(value)), end=' ')


# noinspection PyPep8Naming
def ordered_dump(data, stream=None, Dumper=yaml.Dumper, **keywords):
    """
    writes the dict into an ordered yaml.
    :param data: The ordered dict
    :param stream: the stream
    :param Dumper: the dumper such as yaml.SafeDumper
    """

    # noinspection PyClassHasNoInit
    class OrderedDumper(Dumper):
        """
        A helper class to print an ordered dict
        """
        pass

    def _dict_representer(dumper, data):
        return dumper.represent_mapping(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            data.items())

    OrderedDumper.add_representer(OrderedDict, _dict_representer)
    return yaml.dump(data, stream, OrderedDumper, **keywords)


# noinspection PyPep8Naming
def dprint(OD, mode='dict', s="", indent=' ' * 4, level=0):
    """
    a recursive dict printer method that adds indentations
    TODO: needs better explanation and test example
    :param OD: the ordered dict
    :param mode: the mode is dict
    :param s: TODO
    :param indent: the indentation characters. default is 4
    :param level: the level
    :return: 
    """
    def is_number(s):
        """
        checks if the type of s is a float
        :param s: 
        :return: 
        """
        try:
            float(s)
            return True
        except ValueError:
            return False

    def fstr(s):
        """
        return a string or number
        :param s: teh string or number
        :return: 
        """
        return s if is_number(s) else '"%s"' % s

    if mode != 'dict':
        kv_tpl = '("%s", %s)'
        ST = 'OrderedDict([\n'
        END = '])'
    else:
        kv_tpl = '"%s": %s'
        ST = '{\n'
        END = '}'
    for i, k in enumerate(OD.keys()):
        if type(OD[k]) in [dict, OrderedDict]:
            level += 1
            s += (level - 1) * indent + kv_tpl % (k, ST + dprint(OD[k], mode=mode,
                                                                 indent=indent, level=level) + (
                                                      level - 1) * indent + END)
            level -= 1
        else:
            s += level * indent + kv_tpl % (k, fstr(OD[k]))
        if i != len(OD) - 1:
            s += ","
        s += "\n"
    return s


class Config(object):
    """
    Manage configuration files in yaml format
    """
    @classmethod
    def check_file_for_tabs(cls, filename, verbose=True):
        """identifies if the file contains tabs and returns True if it
        does. It also prints the location of the lines and columns. If
        verbose is set to False, the location is not printed.
        :param filename: the filename
        :type filename: str
        :rtype: True if there are tabs in the file
        """
        filename = path_expand(filename)
        file_contains_tabs = False

        with open(filename, 'r') as f:
            lines = f.read().split("\n")

        line_no = 1
        for line in lines:
            if "\t" in line:
                file_contains_tabs = True
                location = [
                    i for i in range(len(line)) if line.startswith('\t', i)]
                if verbose:
                    print("Tab found in line", line_no, "and column(s)",
                          location)
            line_no += 1
        return file_contains_tabs

    @classmethod
    def path_expand(cls, path):
        """
        expands the path while replacing environment variables, ./, and ~/
        :param path: the path to be expanded
        :type path: string
        :return:the new path
        :rtype: string
        """
        current_dir = "." + os.path.sep
        if path.startswith(current_dir):
            cwd = str(os.getcwd())
            path = path.replace(current_dir, cwd, 1)
        location = os.path.expandvars(os.path.expanduser(path))
        return location

    @classmethod
    def find_file(cls, filename, load_order=None, verbose=False):
        """
        find the specified file in the list of directories that are given in the
        array load_order
        :param filename: the file name
        :type filename: str
        :param load_order: an array with path names in with the filename is looked for.
        :type load_order: list of str
        :param verbose:
        :type verbose: bool
        :return: file name if successful
        :rtype: string if the file exists or None otherwise
        """
        if load_order is None:
            load_order = [".", os.path.join("~", ".cloudmesh")]
        for path in load_order:
            name = Config.path_expand(path + os.path.sep + filename)
            if verbose:
                print("try finding file", name)
            if os.path.isfile(name):
                if verbose:
                    print("Found File", name)
                return name
        return None


class ConfigDict(object):
    """
    Manages the content of cloudmesh.yaml as dict
    """
    __shared_state = {}
    versions = ['4.1']
    data = {}
    filename = None

    def __init__(self,
                 filename,
                 load_order=None,
                 verbose=False,
                 etc=False,
                 reload=False):

        """
        Creates a dictionary from a yaml configuration file
        while using the filename to load it in the specified load_order.
        The load order is an array of paths in which the file is searched.
        By default the load order is set to . and ~/.cloudmesh
        :param filename: the filename
        :type filename: string
        :param load_order: an array with path names in with the filename is looked for.
        :type load_order: list of strings
        :return: an instance of ConfigDict
        :rtype: ConfigDict
        """

        self.__dict__ = self.__shared_state

        if ConfigDict.data != {} and not reload:
            return

        # print ("INIT CONFIGDICT", filename, load_order, verbose, etc)
        self.data = None
        if etc:
            # import cloudmesh_client.etc
            # load_order = [os.path.dirname(cloudmesh_client.etc.__file__)]
            print("etc not supported")
            sys.exit(1)

        if load_order is None:
            self.load_order = [".", os.path.join("~", ".cloudmesh")]
        else:
            self.load_order = load_order

        for path in self.load_order:
            name = Config.path_expand(os.path.join(path, filename))
            if verbose:
                print("try Loading ConfigDict", name)
            if os.path.isfile(name):
                if verbose:
                    print("Loading ConfigDict", name)
                self.load(name)
                ConfigDict.filename = name
                ConfigDict.data = self.data
                return

        # Create default yaml file
        raise ValueError("Could not find file {:} in {:}".format(filename, self.load_order))

    def load(self, filename):
        """
        loads the configuration from the yaml filename
        :param filename:
        :type filename: string
        :return:
        """
        # print ("LOAD CONFIGDICT", filename)

        self.data = BaseConfigDict(filename=Config.path_expand(filename))
        try:
            version = str(self.data["meta"]["version"])
            if version not in self.versions:
                Console.error("The yaml file version must be {}".format(', '.join(self.versions)))
                sys.exit(1)
        except Exception as e:
            Console.error("Your yaml file ~/.cloudmesh/cloudmesh.yaml is not up to date.", traceflag=False)
            Console.error(e.message, traceflag=False)
            sys.exit(1)
            # return self.data

    def write(self, filename=None, output="dict", attribute_indent=4):
        """
        This method writes the dict into various output formats. This includes a dict,
        json, and yaml
        :param filename: the file in which the dict is written
        :param output: is a string that is either "dict", "json", "yaml"
        :param attribute_indent: character indentation of nested attributes in
        """
        if filename is not None:
            location = path_expand(filename)
        else:
            location = self['meta']['location']

            # with open('data.yml', 'w') as outfile:
            #    outfile.write( yaml.dump(data, default_flow_style=True) )

        # Make a backup
        self.make_a_copy(location)

        f = os.open(location, os.O_CREAT | os.O_TRUNC |
                    os.O_WRONLY, stat.S_IRUSR | stat.S_IWUSR)
        if output == "json":
            os.write(f, self.json())
        elif output in ['yml', 'yaml']:
            # d = dict(self)
            # os.write(f, yaml.dump(d, default_flow_style=False))
            os.write(f, ordered_dump(OrderedDict(self),
                                     Dumper=yaml.SafeDumper,
                                     default_flow_style=False,
                                     indent=attribute_indent))
        elif output == "print":
            os.write(f, str(custom_print(self, attribute_indent)))
        else:
            os.write(f, self.dump())
        os.close(f)

    def make_a_copy(self, location=None):
        """
        Creates a backup of the file specified in the location. The backup
        filename  appends a .bak.NO where number is a number that is not yet
        used in the backup directory.
        TODO: This function should be moved to another file maybe XShell
        :param location: the location of the file to be backed up
        """
        import shutil
        destination = backup_name(location)
        shutil.copyfile(location, destination)

    def save(self, filename=None):
        """
        saves the configuration in the given filename,
        if it is none the filename at load time is used.
        :param filename: the file name
        :type filename: string
        :return:
        """
        content = self.data.yaml()
        with open(Config.path_expand(ConfigDict.filename), 'w') as f:
            f.write(content)

    def __setitem__(self, item, value):
        """
        sets an item with the given value while using . formatted keys
        set('a.b.c", value)
        :param item:
        :type item:
        :param value:
        :type value:
        :return:
        """
        keys = None
        if "." in item:
            keys = item.split(".")
        else:
            element = self.data[item]

        element = self.data[keys[0]]
        for key in keys[1:]:
            element = element[key]
        element = value

    def __getitem__(self, item):
        """
        gets an item form the dict. The key is . separated
        use it as follows get("a.b.c")
        :param item:
        :type item:
        :return:
        """
        if "." in item:
            keys = item.split(".")
        else:
            return self.data[item]
        element = self.data[keys[0]]
        for key in keys[1:]:
            element = element[key]
        return element

    def __str__(self):
        """
        returns the dict in yaml format
        :return: returns the yaml output of the dict
        :rtype: string
        """
        return self.data.yaml()

    @property
    def yaml(self):
        """
        returns the dict in yaml format
        :return: returns the yaml output of the dict
        :rtype: string:
        """
        return self.data.yaml()

    def info(self):
        """
        prints out the dict type and its content
        """
        print(type(self.data))
        print(self.data)

    @property
    def json(self, start=None):
        """
        :param start: start key in dot notation
        returns the dict in json format
        :return: json string version
        :rtype: string
        """
        if start is not None:
            data = self.data[start]
        return json.dumps(self.data, indent=4)

    @classmethod
    def check(cls, filename):
        """
        checks the filename if it is syntactically correct and does not include tabs
        :param filename:
        :type filename: string
        :return:
        """
        TODO.implement()

    # noinspection PyPep8Naming
    @classmethod
    def getUser(cls, cloud):
        """
        gets the username for a specified cloud.
        TODO: works currently only for opensatck.
        :param cloud: the name of the cloud
        :return: 
        """
        try:
            config = d = ConfigDict("cloudmesh.yaml")

            d = ConfigDict("cloudmesh.yaml")

            #
            # bug: cloud is none when adding a group
            #

            config = d["cloudmesh"]["clouds"][cloud]
            credentials = config["credentials"]
            cloud_type = config["cm_type"]

            if cloud_type == "openstack":
                return credentials["OS_USERNAME"]
            else:
                raise ValueError("getUser for this cloud type not yet "
                                 "supported: {}".format(cloud))
        except Exception as ex:
            Console.error("problem getting user")


# noinspection PyPep8Naming
def Username():
    """
    returns the username as defined in the profile
    :return: 
    """
    d = ConfigDict("cloudmesh.yaml")

    if "user" not in d["cloudmesh"]["profile"]:
        raise RuntimeError("Profile username is not set in yaml file.")

    user = d["cloudmesh"]["profile"]["user"]
    return user


def main():
    """
    TODO: A test which should actually be moved into a nosetest
    :return: 
    """
    d = ConfigDict("cloudmesh.yaml")
    print(d, end='')
    d.info()

    print(d["meta"])
    print(d["meta.kind"])
    print(d["meta"]["kind"])

    # this does not yet work
    d.data["cloudmesh"]["profile"]["firstname"] = 'ABC'
    print(d)
    d.save()

    import os

    os.system("cat cmd3.yaml")

    print(d.json)
    print(d.filename)
    print("YAML")
    print(d.yaml)


if __name__ == "__main__":
    main()



""" run with

python setup.py install; nosetests -v --nocapture tests/cm_basic/test_shell.py:Test_shell.test_001

nosetests -v --nocapture tests/cm_basic/test_shell.py

or

nosetests -v tests/cm_basic/test_shell.py

"""
from __future__ import print_function

import getpass

from cloudmesh.common.Shell import Shell
from cloudmesh.common.util import HEADING


def run(command):
    parameter = command.split(" ")
    shell_command = parameter[0]
    args = parameter[1:]
    result = Shell.execute(shell_command, args)
    return str(result)


# noinspection PyMethodMayBeStatic,PyPep8Naming
class Test_shell(object):
    """

    """

    def setup(self):
        pass

    def test_001(self):
        HEADING("check if we can run help:return: ")
        r = run("whoami")
        print(r)
        assert getpass.getuser() == r



from hostlist import expand_hostlist


class Parameter(object):
    @classmethod
    def expand(cls, parameter, allow_duplicates=False, sort=False):
        return expand_hostlist(parameter, allow_duplicates=False, sort=False)



from __future__ import print_function

from cloudmesh.db.strdb import YamlDB
from cloudmesh.common.util import path_expand


class Default(object):
    def _index(self, context, key):
        return str(context) + "," + str(key)

    def __init__(self, filename=None):
        if filename is None:
            self.filename = path_expand("~/.cloudmesh/default-data")

        self.data = YamlDB(self.filename)

    def __getitem__(self, context_key):
        try:
            if type(context_key) == tuple:
                context, key = context_key
                index = self._index(context, key)
                if index not in self.data:
                    return None
                else:
                    return self.data[index]
            else:
                d = self.__dict__()
                if context_key not in d:
                    return None
                else:
                    return self.__dict__()[context_key]
        except:
            return None

    def __setitem__(self, context_key, value):
        context, key = context_key
        self.data[self._index(context, key)] = value

    def __delitem__(self, context_key):
        print("DEL")
        if type(context_key) == tuple:
            context, key = context_key
            del self.data[self._index(context, key)]
        else:
            context = context_key
            for element in self.data:
                print("E", element, context)
                if element.startswith(context + ","):
                    del self.data[element]

    def __contains__(self, item):
        for key in self.data:
            if item == self.data[key]:
                return True
        return False

    def __str__(self):
        d = {}
        for element in self.data:
            context, key = element.split(",")
            value = self.data[element]
            if context not in d:
                d[context] = {}

            d[context][key] = value
        # return (str(self.data))
        return str(self.__dict__())

    def __dict__(self):
        d = {}
        for element in self.data:
            context, key = element.split(",")
            value = self.data[element]
            if context not in d:
                d[context] = {}

            d[context][key] = value
        return d

    def __repr__(self):
        return str(self.data)

    def __len__(self):
        return len(self.data)

    # def __add__(self, directory):
    #    for key in directory:
    #        self.data[key] = directory[key]

    # def __sub__(self, keys):
    #    for key in keys:
    #        del self.data[key]

    def close(self):
        self.data.close()


if __name__ == "__main__":
    v = Default()
    print(v)
    v["kilo", "gregor"] = "value"

    assert "value" in v
    del v["kilo", "gregor"]
    # assert "gregor" not in v

    v["kilo", "image"] = "i_k"
    v["kilo", "flavor"] = "f_k"

    v["chameleon", "image"] = "i_c"
    v["chameleon", "flavor"] = "f_c"

    print(v)

    print(v.__repr__())

    print(v["chameleon", "bla"])
    assert v["chameleon", "bla"] is None

    print(v["chameleon"])
    assert v["chameleon"]['image'] == 'i_c'
    print(v["bla"])
    assert v["bla"] is None



"""
authorized key management.
"""
# TODO: needs nosetests
import io
import itertools
import os.path

from six import itervalues

from cloudmesh.common.Shell import Subprocess
from cloudmesh.common.util import tempdir


# TODO:  use our simple subprocess wrapper ?

def get_fingerprint_from_public_key(pubkey):
    """Generate the fingerprint of a public key

    :param str pubkey: the value of the public key
    :returns: fingerprint
    :rtype: str
    """

    # TODO: why is there a tmpdir?
    with tempdir() as workdir:
        key = os.path.join(workdir, 'key.pub')
        with open(key, 'w') as fd:
            fd.write(pubkey)

        cmd = [
            'ssh-keygen',
            '-l',
            '-f', key,
        ]

        p = Subprocess(cmd)
        output = p.stdout.strip()
        bits, fingerprint, _ = output.split(' ', 2)
        return fingerprint


class AuthorizedKeys(object):
    """
    Class to manage authorized keys.
    """
    def __init__(self):
        self._order = dict()
        self._keys = dict()

    @classmethod
    def load(cls, path):
        """
        load the keys from a path
        
        :param path: the filename (path) in which we find the keys
        :return: 
        """
        auth = cls()
        with open(path) as fd:
            for pubkey in itertools.imap(str.strip, fd):
                # skip empty lines
                if not pubkey:
                    continue
                auth.add(pubkey)
        return auth

    def add(self, pubkey):
        """
        add a public key.
        :param pubkey: the filename to the public key
        :return: 
        """
        f = get_fingerprint_from_public_key(pubkey)
        if f not in self._keys:
            self._order[len(self._keys)] = f
            self._keys[f] = pubkey

    def remove(self, pubkey):
        """
        Removes the public key
        TODO: this method is not implemented
        :param pubkey: the filename of the public key
        :return: 
        """
        raise NotImplementedError()

    def __str__(self):

        sio = io.StringIO()

        # TODO: make python 2 and 3 compatible
        # old: for fingerprint in self._order.itervalues():

        for fingerprint in itervalues(self._order):
            key = self._keys[fingerprint]
            sio.write(key)
            sio.write('\n')

        text = sio.getvalue()
        sio.close()

        return text.strip()


if __name__ == '__main__':
    import sys

    path = sys.argv[1]
    auth = AuthorizedKeys.load(path)
    print(auth)



""" run with

python setup.py install; nosetests -v --nocapture tests/cm_basic/test_configdict.py:Test_configdict.test_001

nosetests -v --nocapture tests/cm_basic/test_configdict.py

or

nosetests -v tests/cm_basic/test_configdict.py

"""
from __future__ import print_function

import os
import shutil

from cloudmesh.common.ConfigDict import ConfigDict
from cloudmesh.common.util import HEADING


# noinspection PyMethodMayBeStatic,PyMethodMayBeStatic,PyMethodMayBeStatic,PyPep8Naming,PyBroadException,PyBroadException
class Test_configdict:
    root_path = os.path.abspath(os.sep)
    cwd_path = os.getcwd()

    def setup(self):
        # os.system("cm help")
        # self.etc_yaml = os.path.join(self.cwd_path, "cloudmesh_client", "etc", "cloudmesh.yaml")
        self.tmp_yaml = os.path.join(self.root_path, "tmp", "cloudmesh.yaml")
        self.tmp_dir = os.path.join(self.root_path, "tmp")
        pass

    # noinspection PyPep8Naming
    def tearDown(self):
        pass

    def test_001_read(self):
        HEADING("test if cloudmesh.yaml is loaded")
        d = ConfigDict("cloudmesh.yaml",
                       verbose=True)

        assert d["cloudmesh"]["profile"]["firstname"] != ""
        assert len(d["cloudmesh"]["clouds"]) > 0

        """
        #
        # DO NOT DO THIS TEST AS LOGIC OF AUTOCREATION HAS CHANGED
        #
        # try:
        #     d = ConfigDict("cloudmesh.yam",
        #                    verbose=True)
        #     print("the file cloudmesh.yam should not exists")
        #     assert False
        # except Exception as e:
        #     print (">>>>>>>", e)
        #     assert str(e).startswith("Could not find")
        """

    """ do not test for etc
    def test_002_set(self):
        HEADING("testing to set a value in the dict")
        shutil.copy(self.etc_yaml, self.tmp_yaml)
        d = ConfigDict("cloudmesh.yaml",
                       load_order=[self.tmp_dir],
                       verbose=True)
        d["cloudmesh"]["profile"]["firstname"] = "Gregor"
        d.save()

        d = ConfigDict("cloudmesh.yaml",
                       load_order=[self.tmp_dir],
                       verbose=True)
        assert d["cloudmesh"]["profile"]["firstname"] == "Gregor"
    """
        
    def test_003_json(self):
        HEADING("test if json is produced")
        d = ConfigDict("cloudmesh.yaml",
                       verbose=True)

        assert d.json.startswith('{')

        try:
            assert not isinstance(d.json, str)
            print("json should be string")
            assert False
        except Exception as e:
            assert isinstance(d.json, str)

    def test_004_yaml(self):

        HEADING("test if yaml is produced")
        d = ConfigDict("cloudmesh.yaml",
                       verbose=True)
        result = d.yaml

        try:
            assert result.startswith("meta")
        except Exception as e:
            print("not valid yaml file.")
            assert False


"""	def main():
    d = ConfigDict("cmd3.yaml")
    print (d, end='')
    print(Printer.write(self.cm.info()))


    print (d["meta"])
    print (d["meta.kind"])
    print (d["meta"]["kind"])

    # this does not yet work
    print (d)
    d.save()

    import os
    os.system("cat cmd3.yaml")

    print(d.json)
    print(d.filename)

if __name__ == "__main__":
    main()
"""



"""
Class for starting and stopping named timers.

This class is based on a similar java class in cyberaide, and java cog kit.

"""
from __future__ import print_function

import os
import time


class StopWatch(object):
    """
    A class to measure times between events.
    """
    debug = False
    verbose = True
    # Timer start dict
    timer_start = {}
    # Timer end dict
    timer_end = {}
    # Timer diff
    timer_elapsed = {}

    @classmethod
    def keys(cls):
        """returns the names of the timers"""
        return list(cls.timer_end.keys())

    @classmethod
    def start(cls, name):
        """
        starts a timer with the given name.

        :param name: the name of the timer
        :type name: string
        """
        if cls.debug:
            print("Timer", name, "started ...")
        cls.timer_start[name] = time.time()

    @classmethod
    def stop(cls, name):
        """
        stops the timer with a given name.

        :param name: the name of the timer
        :type name: string
        """
        cls.timer_end[name] = time.time()
        if cls.debug:
            print("Timer", name, "stopped ...")

    @classmethod
    def get(cls, name):
        """
        returns the time of the timer.

        :param name: the name of the timer
        :type name: string
        :rtype: the elapsed time
        """
        if name in cls.timer_end:
            cls.timer_elapsed[name] = cls.timer_end[name] - \
                                      cls.timer_start[name]
            return cls.timer_elapsed[name]
        else:
            return "undefined"

    @classmethod
    def clear(cls):
        """
        clear start and end timer_start
        """
        cls.timer_start.clear()
        cls.timer_end.clear()

    @classmethod
    def print(cls, *args):
        """
        prints a timer. The first argument is the label if it exists, the last is the timer
        :param args: label, name 
        :return: 
        """
        if cls.verbose:
            if len(args) == 2:
                print(args[0], str("{0:.2f}".format(cls.get(args[1]))), "s")
            else:
                raise Exception("StopWatch: wrong number of arguments")

    @classmethod
    def __str__(cls):
        s = ""
        for t in cls.timer_end:
            data = {"label": t,
                    "start": str(cls.timer_start[t]),
                    "end": str(cls.timer_end[t]),
                    "elapsed": str(cls.get(t)),
                    "newline": os.linesep}
            s += "{label} {start} {end} {elapsed} {newline}".format(**data)
        return s
